/*
 * DUMP - Make a dump of a Ti-89/Ti-92+/V200 system and output a summary into a file.
 * Copyright (C) 2003 PpHd / Time To Team
 *
 * This program is free software ; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along with this program;
 * if not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#define	DUMP_VERSION 0x0033

#define USE_TI89
#define USE_TI92PLUS
#define USE_V200
#define MIN_AMS 100
#define NO_CALC_DETECT
#define OPTIMIZE_ROM_CALLS 

#include <tigcclib.h>

#define OO_firstACB (*((short*)(_rom_call_addr(451))))
#define OO_NextACB _rom_call(short,(short),402)
#define OO_GetAppAttr _rom_call(void*,(short,long),3FC)
#define OO_GetAppName(obj) (unsigned char *)OO_GetAppAttr(obj,2)

/*
 * What to save ?
 *	+ VECTORS
 *		VECTOR Num / addr.l / dump the first 8 bytes.
 *	+ VAT :
 *		NAME + Flags + Handle + Type
 *	+ HANDLES:
 *		Hd / Addr / Size / Dump first 4, last 4
 *	+ HEAP:
 *		FreeHandles / HeapMax / HeapAvail
 *	+ EV_hook : 
 *		Go throught the linked list / addr / Dump the first 8 bytes
 *	+ BOOT PARM CODE : const void *FL_getHardwareParmBlock();
 *	+ AMS VERSION	 : VERSION NUMBER / unsigned short FL_getVerNum (void);
 *	+ HW2PATCH	 : YES / NO
 *	+ MAXMEM	 : YES / NO
 *	+ HW2TSR	 : YES / NO
 *	+ ARCHIVE MEMORY : void EM_survey (unsigned long *inUse, unsigned long *freeAfterGC, unsigned long *free, unsigned long *unusedSectors, unsigned long *badSectors, unsigned long *allExceptBaseCode);
 *	+ APPLICATION	 : OO_firstACB / OO_NextACB : ACB (22) + Header (56)
 *	+ END
 * FORMAT:
 *	TYPE (VECTOR, EV_...).b / SIZE (in words).b / DATA
 */
typedef enum {END = 0, VECTOR, EVHK_HOOK, UNK_HOOK, SYM, HANDLES, HEAP, PARM_BLOCK, AMS, PATCH, ARCHIVE, APPS, DUMP, OTH_EXT, KERNEL_EXT, PACKARC_EXT, PACKARC_FILE, DUMP_MAX } 
	DUMP_TYPE;
	
HANDLE	h_file;
short	*file;
short	*end;

void	SaveFile()
{
	// Finish File format
	*(file++) = 0;				// End of DUMP
	*(file++) = 'D'*256 + 'U';
	*(file++) = 'M'*256 + 'P';
	*(file++) = 0*256 + 0xF8;

	// Resize handle
	short	*start = HeapDeref(h_file);
	long size = *(start) = (char *) file -  (char *) start - 2;
	HeapUnlock(h_file);
	HeapRealloc(h_file, size+2);

	// Add it to the VAT
	SYM_ENTRY *sym = DerefSym(SymAdd(SYMSTR("dumpfile")));
	if (sym == NULL)
		{
		ST_helpMsg("Can't add 'dumpfile' to the VAT.");
		HeapFree(h_file);
		exit(1);
		}
	sym->handle = h_file;
}

void	AddType(DUMP_TYPE t, short size)
{
	if (size >= 256 || t >= DUMP_MAX)
		{
		ST_helpMsg("Type's size too big !");
		HeapFree(h_file);
		exit(2);
		}
	if (file + size > end)
		{
		ST_helpMsg("Memory overflow. Partial Dump.");
		SaveFile();
		exit(3);
		}
	*(file++) = t * 256 + size;
}

BOOL h220xtsrInstalled(void)
{		
	switch(*(ULONG*)(*(ULONG*)0xAC-8))
	{
		case 0x32547372:	//2Tsr (new signature)
		case 0x32545352:	//2TSR (old signature)
			return TRUE;
		break;
		
		default:
			return FALSE;		
	}
}

unsigned long GetHardwareVersion()
{
  unsigned long hwpb, *rombase;
  rombase = (unsigned long *)((*(unsigned long *)0xC8) & 0x600000);
  hwpb = rombase[65];
  return (hwpb - (unsigned long)rombase < 0x10000 &&
    *(unsigned short *)hwpb > 22 ? *(unsigned long *)(hwpb + 22) : 1);
}

UCHAR HW2PatchInstalled()
{
  ULONG *ptr;
  if (*(ULONG *)0xAC == 0x100)
    return 1;
  ptr = (ULONG *)(((void *)EX_stoBCD)+0x52);
  if (*ptr++ != 0x700000)
    return 1;
  if (*ptr != 0x36BCFFDF)
    return 1;
  return 0;
} 

void	_main(void)
{
	unsigned int i,j;
	HeapCompress();

	unsigned short	freehd = FreeHandles();
	unsigned long	avail = HeapAvail();
	unsigned long	max = HeapMax();

	h_file = HeapAlloc(max);
	file = HLock(h_file);
	end = file + (max/2) - 7 - 2;
	file++;				// Size will be filled after

	AddType(DUMP, 1);
	*(file++) = DUMP_VERSION;
	
	AddType(AMS, (40+12)/2);
	AB_prodname((char *) file);	file+=20;
	AB_prodid((char *) file);	file+=6;

	AddType(HEAP, 5);
	*(file++) = freehd;
	*(file++) = avail>>16; *(file++) = avail&0xFFFF;
	*(file++) = max>>16; *(file++) = max&0xFFFF;
	
	unsigned char *a0;
	short *data = (short *) FL_getHardwareParmBlock();
	unsigned short len = *(data++)/2;
	AddType(PARM_BLOCK, len);
	for(i = 0 ; i < len ; i++)
		*(file++) = *(data++);
	
	for(i = 0 ; i < 64 ; i++)
		{
		AddType(VECTOR, 1+2+4);
		*(file++) = i;
		data = (short *) (unsigned long) (i*4);
		*(file++) = *(data);
		*(file++) = *(data+1);
		data = (short *) *((long *) data);
		if ((unsigned long) data & 1)
			for(j = 0 ; j < 4 ; j++)
				*(file++) = 0;
		else	for(j = 0 ; j < 4 ; j++)
			 	*(file++) = *(data++);
		}
	
	HANDLE	h;
	for(h = 1 ; h < 2000 ; h++)
		{
		data = HeapDeref(h);
		if (data)
			{
			AddType(HANDLES, 1+2+1+4);
			*(file++) = h;
			*(file++) = ((unsigned long) data) >>16;
			*(file++) = ((unsigned long) data) & 0xFFFF;
			data--; 			// Size !
			for(j = 0 ; j < 5 ; j++)
			 	*(file++) = *(data++);
			}
		}

	SYM_ENTRY *sym = SymFindFirst(NULL, FO_RECURSE);
	do {
		AddType(SYM, 9+5);
		data = (short *) sym;
		strncpy((char*) file, SymFindFolderName(), 10);
		file+=5;
		for(i = 0 ; i < 7 ; i++)
			*(file++) = *(data++);
		*(file++) = i = *(HToESI(sym->handle));
		a0 = HeapDeref(sym->handle);
		len = *((short *) a0);
		register unsigned short checksum = 0;
		for(j = 0 ; j < len ; j++)
			checksum += *(a0++);
		*(file++) = checksum;
		if (i == 0xF3)		// ASM
			{
			a0 = HeapDeref(sym->handle);
			if (!strncmp(a0+6,"68k",3))	// Kernel File
				{
				AddType(KERNEL_EXT, (2+20)/2);
		                *(file++) = *(unsigned short*)(a0+18); // Version + Flags
		                if (*(unsigned short*)(a0+12))
		                    strncpy((char *) file, a0 + 2 + *(unsigned short*)(a0+12), 20);
				else	*(file) = 0;
				file +=10;
				}
			else if (!strncmp(a0+4,"68cA",4))	// Pack Archive
				{
				AddType(PACKARC_EXT, (4+8+2)/2);
				memcpy((unsigned char *) file, a0+18, 14);
			        file+=7;
			        {
			         char *fi = a0 + 21;
			         short nbr = *((unsigned char *) a0+18);
			         for(  ; nbr > 0 ; nbr--) {
			         	while (*(fi++));	// Skip this entry (skip the first one too)
					AddType(PACKARC_FILE, (4+8+2)/2);
					memcpy((char *) file, fi, 14);
					file+=7;
					}
				}
				}
			}
		else if (i == 0xF8)		// OTHER
			{
			AddType(OTH_EXT, 4);
			a0 = HToESI(sym->handle);
			a0--; //0
			while (*(--a0));
			strncpy((char *) file, a0+1, 8);
			file +=4;
			}
		sym = SymFindNext();
	} while (sym);
	
	AddType(ARCHIVE, 12);
	EM_survey((unsigned long *) file, (unsigned long *) (file+2), (unsigned long *) (file+4),
		 (unsigned long *) (file+6), (unsigned long *) (file+8), (unsigned long *) (file+10));
	file += 12;

	AddType(PATCH, 3);
	*(file++) = GetHardwareVersion();
	*(file++) = HW2PatchInstalled();
	*(file++) = h220xtsrInstalled();
	// How to detect MaxMem ?
	
	unsigned long	*d = (unsigned long *) EV_hook;
	while (d)
		{
		if (d[-4] == 0x6576486B /*'evHk'*/)
			{
			AddType(EVHK_HOOK, 2+10);
			*(file++) = (short) ((unsigned long) d>>16);
			*(file++) = (short) (unsigned long) (d);
			data = (short *) &d[-3];
			for(j = 0 ; j < 10 ; j++)	// Name (8 + 4 Old + 8 1ers dump)
				*(file++) = *(data++);
			d = (unsigned long *) d[-1];
			if ((unsigned long) d & 1)
				d = NULL;
			}	
		else	{
			AddType(UNK_HOOK, 2+4);
			*(file++) = (short) ((unsigned long) d>>16);
			*(file++) = (short) (unsigned long) (d);
			data = (short *) d;
			for(j = 0 ; j < 4 ; j++)	// 8 1ers dump
				*(file++) = *(data++);
			d = NULL;
			}
		}
	
	if (AMS_2xx)
		{
		h = OO_firstACB;
		while (h)
			{
			AddType(APPS, (20)/2);
			strncpy((char *) file, OO_GetAppName(h), 20);
			file+=10;
			h = OO_NextACB(h);
			}
		}

	SaveFile();
	ST_helpMsg("Dump system done.");
}
